#!/usr/bin/env python
# Stegano - Stegano is a pure Python steganography module.
# Copyright (C) 2010-2026 Cédric Bonhomme - https://www.cedricbonhomme.org
#
# For more information : https://github.com/cedricbonhomme/Stegano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

__author__ = "Cedric Bonhomme"
__version__ = "$Revision: 0.9.4 $"
__date__ = "$Date: 2010/10/01 $"
__revision__ = "$Date: 2019/06/06 $"
__license__ = "GPLv3"

from typing import cast

from PIL import Image


def steganalyse(img: Image.Image) -> Image.Image:
    """
    Steganlysis of the LSB technique.
    """
    encoded = Image.new(img.mode, (img.size))
    width, height = img.size
    for row in range(height):
        for col in range(width):
            if pixel := cast(tuple[int, int, int], img.getpixel((col, row))):
                r, g, b = pixel[0:3]
            else:
                raise Exception("Error during steganlysis.")
            if r % 2 == 0:
                r = 0
            else:
                r = 255
            if g % 2 == 0:
                g = 0
            else:
                g = 255
            if b % 2 == 0:
                b = 0
            else:
                b = 255
            encoded.putpixel((col, row), (r, g, b))
    return encoded



#!/usr/bin/env python
# Stegano - Stegano is a pure Python steganography module.
# Copyright (C) 2010-2026 Cédric Bonhomme - https://www.cedricbonhomme.org
#
# For more information : https://github.com/cedricbonhomme/Stegano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

__author__ = "Cedric Bonhomme"
__version__ = "$Revision: 0.7 $"
__date__ = "$Date: 2016/03/18 $"
__revision__ = "$Date: 2019/06/04 $"
__license__ = "GPLv3"

import inspect

import crayons

try:
    from stegano import lsb
    from stegano.lsb import generators
except Exception:
    print("Install stegano: pipx install Stegano")

import argparse

from stegano import tools


class ValidateGenerator(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        valid_generators = [
            generator[0]
            for generator in inspect.getmembers(generators, inspect.isfunction)
        ]
        # Verify that the generator is valid
        generator = values[0]
        if generator not in valid_generators:
            raise ValueError("Unknown generator: %s" % generator)
        # Set the generator_function arg of the parser
        setattr(args, self.dest, values)


def main():
    parser = argparse.ArgumentParser(prog="stegano-lsb")
    subparsers = parser.add_subparsers(
        help="sub-command help", dest="command", required=True
    )

    # Subparser: Hide
    parser_hide = subparsers.add_parser("hide", help="hide help")
    # Original image
    parser_hide.add_argument(
        "-i",
        "--input",
        dest="input_image_file",
        required=True,
        help="Input image file.",
    )
    parser_hide.add_argument(
        "-e",
        "--encoding",
        dest="encoding",
        choices=tools.ENCODINGS.keys(),
        default="UTF-8",
        help="Specify the encoding of the message to hide."
        " UTF-8 (default) or UTF-32LE.",
    )

    # Generator
    parser_hide.add_argument(
        "-g",
        "--generator",
        dest="generator_function",
        action=ValidateGenerator,
        nargs="*",
        required=False,
        default=None,
        help="Generator (with optional arguments)",
    )

    # Shift the message to hide
    parser_hide.add_argument(
        "-s", "--shift", dest="shift", default=0, help="Shift for the generator"
    )

    group_secret = parser_hide.add_mutually_exclusive_group(required=True)
    # Non binary secret message to hide
    group_secret.add_argument(
        "-m", dest="secret_message", help="Your secret message to hide (non binary)."
    )
    # Binary secret message to hide
    group_secret.add_argument(
        "-f", dest="secret_file", help="Your secret to hide (Text or any binary file)."
    )

    # Image containing the secret
    parser_hide.add_argument(
        "-o",
        "--output",
        dest="output_image_file",
        required=True,
        help="Output image containing the secret.",
    )

    # Subparser: Reveal
    parser_reveal = subparsers.add_parser("reveal", help="reveal help")
    parser_reveal.add_argument(
        "-i",
        "--input",
        dest="input_image_file",
        required=True,
        help="Input image file.",
    )
    parser_reveal.add_argument(
        "-e",
        "--encoding",
        dest="encoding",
        choices=tools.ENCODINGS.keys(),
        default="UTF-8",
        help="Specify the encoding of the message to reveal."
        " UTF-8 (default) or UTF-32LE.",
    )

    # Generator
    parser_reveal.add_argument(
        "-g",
        "--generator",
        dest="generator_function",
        action=ValidateGenerator,
        nargs="*",
        required=False,
        help="Generator (with optional arguments)",
    )

    # Shift the message to reveal
    parser_reveal.add_argument(
        "-s", "--shift", dest="shift", default=0, help="Shift for the generator"
    )
    parser_reveal.add_argument(
        "-o",
        dest="secret_binary",
        help="Output for the binary secret (Text or any binary file).",
    )

    # Subparser: List generators
    subparsers.add_parser("list-generators", help="list-generators help")

    arguments = parser.parse_args()

    if arguments.command != "list-generators":
        if not arguments.generator_function:
            generator = None
        else:
            try:
                if arguments.generator_function[0] == "LFSR":
                    # Compute the size of the image for use by the LFSR generator if needed
                    tmp = tools.open_image(arguments.input_image_file)
                    size = tmp.width * tmp.height
                    tmp.close()
                    arguments.generator_function.append(size)
                if len(arguments.generator_function) > 1:
                    generator = getattr(generators, arguments.generator_function[0])(
                        *[int(e) for e in arguments.generator_function[1:]]
                    )
                else:
                    generator = getattr(generators, arguments.generator_function[0])()

            except AttributeError:
                print(f"Unknown generator: {arguments.generator_function}")
                exit(1)

    if arguments.command == "hide":
        if arguments.secret_message is not None:
            secret = arguments.secret_message
        elif arguments.secret_file != "":
            secret = tools.binary2base64(arguments.secret_file)

        img_encoded = lsb.hide(
            image=arguments.input_image_file,
            message=secret,
            generator=generator,
            shift=int(arguments.shift),
            encoding=arguments.encoding,
        )
        try:
            img_encoded.save(arguments.output_image_file)
        except Exception as e:
            # If hide() returns an error (Too long message).
            print(e)

    elif arguments.command == "reveal":
        try:
            secret = lsb.reveal(
                encoded_image=arguments.input_image_file,
                generator=generator,
                shift=int(arguments.shift),
                encoding=arguments.encoding,
            )
        except IndexError:
            print("Impossible to detect message.")
            exit(0)
        if arguments.secret_binary is not None:
            data = tools.base642binary(secret)
            with open(arguments.secret_binary, "wb") as f:
                f.write(data)
        else:
            print(secret)

    elif arguments.command == "list-generators":
        all_generators = inspect.getmembers(generators, inspect.isfunction)
        for generator in all_generators:
            print("Generator id:")
            print(f"    {crayons.green(generator[0], bold=True)}")
            print("Desciption:")
            print(f"    {generator[1].__doc__}")



#!/usr/bin/env python
# Stegano - Stegano is a pure Python steganography module.
# Copyright (C) 2010-2026 Cédric Bonhomme - https://www.cedricbonhomme.org
#
# For more information : https://github.com/cedricbonhomme/Stegano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

__author__ = "Cedric Bonhomme"
__version__ = "$Revision: 0.2 $"
__date__ = "$Date: 2010/10/01 $"
__revision__ = "$Date: 2021/11/01 $"
__license__ = "GPLv3"

import typing
from collections import Counter, OrderedDict


def steganalyse(img):
    """
    Steganlysis of the LSB technique.
    """
    width, height = img.size
    colours_counter: typing.Counter[int] = Counter()
    for row in range(height):
        for col in range(width):
            r, g, b = img.getpixel((col, row))
            colours_counter[r] += 1

    most_common = colours_counter.most_common(10)
    dict_colours = OrderedDict(
        sorted(list(colours_counter.items()), key=lambda t: t[1])
    )

    colours: float = 0
    for colour in list(dict_colours.keys()):
        colours += colour
    colours = colours / len(dict_colours)

    # return colours.most_common(10)
    return list(dict_colours.keys())[:30], most_common



#!/usr/bin/env python
# Stegano - Stegano is a pure Python steganography module.
# Copyright (C) 2010-2026 Cédric Bonhomme - https://www.cedricbonhomme.org
#
# For more information : https://github.com/cedricbonhomme/Stegano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

__author__ = "Cedric Bonhomme"
__version__ = "$Revision: 0.7 $"
__date__ = "$Date: 2016/03/13 $"
__revision__ = "$Date: 2019/05/31 $"
__license__ = "GPLv3"

from typing import IO, Iterator, Union

from stegano import tools

from .generators import identity


def hide(
    image: Union[str, IO[bytes]],
    message: str,
    generator: Union[None, Iterator[int]] = None,
    shift: int = 0,
    encoding: str = "UTF-8",
    auto_convert_rgb: bool = False,
):
    """Hide a message (string) in an image with the
    LSB (Least Significant Bit) technique.
    """
    hider = tools.Hider(image, message, encoding, auto_convert_rgb)
    width = hider.encoded_image.width

    if not generator:
        generator = identity()

    while shift != 0:
        next(generator)
        shift -= 1

    while hider.encode_another_pixel():
        generated_number = next(generator)

        col = generated_number % width
        row = int(generated_number / width)

        hider.encode_pixel((col, row))

    return hider.encoded_image


def reveal(
    encoded_image: Union[str, IO[bytes]],
    generator: Union[None, Iterator[int]] = None,
    shift: int = 0,
    encoding: str = "UTF-8",
    close_file: bool = True,
):
    """Find a message in an image (with the LSB technique)."""
    revealer = tools.Revealer(encoded_image, encoding, close_file)
    width = revealer.encoded_image.width

    if not generator:
        generator = identity()

    while shift != 0:
        next(generator)
        shift -= 1

    while True:
        generated_number = next(generator)

        col = generated_number % width
        row = int(generated_number / width)

        if revealer.decode_pixel((col, row)):
            return revealer.secret_message



#!/usr/bin/env python
# Stegano - Stegano is a pure Python steganography module.
# Copyright (C) 2010-2026 Cédric Bonhomme - https://www.cedricbonhomme.org
#
# For more information : https://github.com/cedricbonhomme/Stegano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

__author__ = "Cedric Bonhomme"
__version__ = "$Revision: 0.3 $"
__date__ = "$Date: 2010/10/01 $"
__revision__ = "$Date: 2017/05/04 $"
__license__ = "GPLv3"

import base64
import itertools
from functools import reduce
from typing import IO, List, Union, cast

from PIL import Image

ENCODINGS = {"UTF-8": 8, "UTF-32LE": 32}


def a2bits(chars: str) -> str:
    """Converts a string to its bits representation as a string of 0's and 1's.

    >>> a2bits("Hello World!")
    '010010000110010101101100011011000110111100100000010101110110111101110010011011000110010000100001'
    """
    return bin(reduce(lambda x, y: (x << 8) + y, (ord(c) for c in chars), 1))[3:]


def a2bits_list(chars: str, encoding: str = "UTF-8") -> List[str]:
    """Convert a string to its bits representation as a list of 0's and 1's.

    >>>  a2bits_list("Hello World!")
    ['01001000',
    '01100101',
    '01101100',
    '01101100',
    '01101111',
    '00100000',
    '01010111',
    '01101111',
    '01110010',
    '01101100',
    '01100100',
    '00100001']
    >>> "".join(a2bits_list("Hello World!"))
    '010010000110010101101100011011000110111100100000010101110110111101110010011011000110010000100001'
    """
    return [bin(ord(x))[2:].rjust(ENCODINGS[encoding], "0") for x in chars]


def bs(s: int) -> str:
    """Converts an int to its bits representation as a string of 0's and 1's."""
    return str(s) if s <= 1 else bs(s >> 1) + str(s & 1)


def setlsb(component: int, bit: str) -> int:
    """Set Least Significant Bit of a colour component."""
    return component & ~1 | int(bit)


def n_at_a_time(items: List[int], n: int, fillvalue: str):
    """Returns an iterator which groups n items at a time.
    Any final partial tuple will be padded with the fillvalue

    >>> list(n_at_a_time([1, 2, 3, 4, 5], 2, 'X'))
    [(1, 2), (3, 4), (5, 'X')]
    """
    it = iter(items)
    return itertools.zip_longest(*[it] * n, fillvalue=fillvalue)


def binary2base64(binary_file: str) -> str:
    """Convert a binary file (OGG, executable, etc.) to a
    printable string.
    """
    # Use mode = "rb" to read binary file
    with open(binary_file, "rb") as bin_file:
        encoded_string = base64.b64encode(bin_file.read())
    return encoded_string.decode()


def base642binary(b64_fname: str) -> bytes:
    """Convert a printable string to a binary file."""
    b64_fname += "==="
    return base64.b64decode(b64_fname)


def open_image(fname_or_instance: Union[str, IO[bytes], Image.Image]) -> Image.Image:
    """Opens an image and returns it.

    :param fname_or_instance: Can be a path to the image (str),
                              a file-like object (IO[bytes]),
                              or a PIL Image instance.
    """
    if isinstance(fname_or_instance, Image.Image):
        return fname_or_instance

    return Image.open(fname_or_instance)


class Hider:
    def __init__(
        self,
        input_image: Union[str, IO[bytes]],
        message: str,
        encoding: str = "UTF-8",
        auto_convert_rgb: bool = False,
    ):
        self._index = 0

        message_length = len(message)
        assert message_length != 0, "message length is zero"

        image = open_image(input_image)

        if image.mode not in ["RGB", "RGBA"]:
            if not auto_convert_rgb:
                print(f"The mode of the image is not RGB. Mode is {image.mode}")
                answer = input("Convert the image to RGB ? [Y / n]\n") or "Y"
                if answer.lower() == "n":
                    raise Exception("Not a RGB image.")

            image = image.convert("RGB")

        self.encoded_image = image.copy()
        image.close()

        # Pre-compute pixel access and mode for performance
        self._is_rgba = self.encoded_image.mode == "RGBA"
        self._pixels = self.encoded_image.load()

        message = str(message_length) + ":" + str(message)
        self._message_bits = "".join(a2bits_list(message, encoding))
        self._message_bits += "0" * ((3 - (len(self._message_bits) % 3)) % 3)

        width, height = self.encoded_image.size
        npixels = width * height
        self._len_message_bits = len(self._message_bits)

        if self._len_message_bits > npixels * 3:
            raise Exception(
                f"The message you want to hide is too long: {message_length}"
            )

    def encode_another_pixel(self):
        return True if self._index + 3 <= self._len_message_bits else False

    def encode_pixel(self, coordinate: tuple):
        # Use direct pixel access for performance
        pixel = self._pixels[coordinate]

        if self._is_rgba:
            r, g, b, a = cast(tuple[int, int, int, int], pixel)
        else:
            r, g, b = cast(tuple[int, int, int], pixel)

        # Change the Least Significant Bit of each colour component.
        r = setlsb(r, self._message_bits[self._index])
        g = setlsb(g, self._message_bits[self._index + 1])
        b = setlsb(b, self._message_bits[self._index + 2])

        # Save the new pixel using direct pixel access
        if self._is_rgba:
            self._pixels[coordinate] = (r, g, b, a)
        else:
            self._pixels[coordinate] = (r, g, b)

        self._index += 3


class Revealer:
    def __init__(
        self,
        encoded_image: Union[str, IO[bytes]],
        encoding: str = "UTF-8",
        close_file: bool = True,
    ):
        self.encoded_image = open_image(encoded_image)
        self._encoding_length = ENCODINGS[encoding]
        self._buff, self._count = 0, 0
        self._bitab: List[str] = []
        self._limit: Union[None, int] = None
        self._limit_str = ""  # Accumulator for length prefix to avoid O(n²) joins
        self.secret_message = ""
        self.close_file = close_file

        # Pre-compute pixel access and mode for performance
        self._is_rgba = self.encoded_image.mode == "RGBA"
        self._pixels = self.encoded_image.load()

    def decode_pixel(self, coordinate: tuple):
        # Use direct pixel access for performance
        pixel = cast(
            tuple[int, int, int] | tuple[int, int, int, int],
            self._pixels[coordinate],
        )

        if self._is_rgba:
            pixel = pixel[:3]  # ignore the alpha

        for color in pixel:
            self._buff += (color & 1) << (self._encoding_length - 1 - self._count)
            self._count += 1

            if self._count == self._encoding_length:
                char = chr(self._buff)
                self._bitab.append(char)
                self._buff, self._count = 0, 0

                # Accumulate length prefix incrementally to avoid O(n²) joins
                if self._limit is None:
                    if char == ":":
                        if self._limit_str.isdigit():
                            self._limit = int(self._limit_str)
                        else:
                            raise IndexError("Impossible to detect message.")
                    else:
                        self._limit_str += char

        prefix_len = len(str(self._limit)) + 1
        if len(self._bitab) - prefix_len == self._limit:
            self.secret_message = "".join(self._bitab)[prefix_len:]
            if self.close_file:
                self.encoded_image.close()
            return True
        else:
            return False
